#!/usr/bin/env ruby

"""
	Simple Stats V .01
	===========
		
		Load and parse log files for Simple Thought Productions to generate nice
		output to make sense of the log data.
	
	Usage
	-----
	
		./simplestats.rb filename.log              - Outputs the Video information to the console.
		./simplestats.rb filename.log >today.log   - Outputs the Video information to a file named today.log
	
	License
	-------
		
		Copyright (C) 2007 Daniel G. Taylor

		This program is free software; you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation; either version 2 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program; if not, write to the Free Software
		Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
	
"""

require 'SVG/Graph/Bar'

# This regular expression breaks up a log entry into parts
$log_entry_regex = /(.*) - - \[(.*)\] (.*) : (.*) : (.*)" - "-"(.*)/

#" stop stupid syntax highlighters that don't understand ruby regexp syntax'

# An array to hold all log items
$log = []

$NAME = "Video Stats"
$AUTHOR = "Daniel G. Taylor"
$VERSION = "0.1"
$COPYRIGHT = "2007"

"""
	Add a method to the String class to allow on-demand interpolation. You
	MUST pass a binding() to the call so that variables are found.
"""
class String
	def interpolate(bind)
		eval "<<END_REEVAL\n" + self + "\nEND_REEVAL\n", bind
	end
end

"""
	Reopen the Graph base class and add the ability to draw extra text areas
	so that we can render explanation text into them. This uses flow elements
	from the SVG 1.2 spec so they may or may not show up depending on the
	viewer until SVG 1.2 is finalized.
"""
module SVG
	module Graph
		class Graph
			# Alias the old start_svg function so we can inject out new method
			alias :old_start_svg :start_svg
			
			def start_svg
				old_start_svg
				draw_text
			end
		
			# Add a new text item to be rendered
			def add_text(x, y, w, h, text, styleclass="subTitle")
				if not @extra_text
					@extra_text = []
				end
				@extra_text.push [x, y, w, h, text, styleclass]
			end
			
			# Draw the text elements added above
			def draw_text
				@extra_text.each do |x, y, w, h, text, styleclass|
					root = @root.add_element("flowRoot", {
						"class" => styleclass
					})
					region = root.add_element("flowRegion")
					region.add_element("rect", {
						"x" => x,
						"y" => y,
						"width" => w,
						"height" => h,
						"fill" => "none"
					})
					root.add_element("flowPara").text = text
				end
			end
		end
	end
end

"""
	Define a new method to the base object of all other objects so that
	we get something like named arguments by passing dicts.
"""
class Object
	"""
		Initialize the object with a given dictionary object to simulate
		named arguments.
	"""
	def init_with(config)
		config.each do |key, value|
			self.send(key.to_s + "=", value) if methods.include? key.to_s
		end
	end
end

"""
	An object representing a single entry in the log file, containing
	information about a client's address, video accessed, etc.
"""
class LogItem
	attr_accessor :ip, :date, :name, :filename, :finished, :agent
	
	def initialize(config)
		init_with({
			:ip        => "",
			:date      => "",
			:name      => "",
			:filename  => "",
			:finished  => false,
			:agent     => ""
		})
		init_with config
	end
end

"""
	An object representing a summary item which contains stats on particular
	video files, such as the number of hits and unique visitors.
"""
class SummaryItem
	attr_accessor :name, :hits, :full_views, :unique_ips
	
	def initialize(config)
		init_with({
			:name       => "",
			:hits       => 0,
			:full_views => 0,
			:unique_ips => []
		})
		init_with config
	end
	
	def to_s
		@name.to_s + " : " + @hits.to_s + " Starts : " + @full_views.to_s + \
			" Stops : " + @unique_ips.length.to_s + " Unique IPs"
	end
end

"""
	Read each item in the log file and store it in a useful format. Each entry
	is stored as a LogItem object.
		  
"""
def read_log(filename)
	cur = 1
	File.open(filename, "r") do |f|
		while line = f.gets
			begin
				m = $log_entry_regex.match line
				name = m[3][1..-1] == "" ? "No name" : m[3][1..-1]
				if m[5].strip != "stop"
					entry = LogItem.new({
						:ip       => m[1],
						:date     => m[2],
						:name     => name,
						:filename => m[4],
						:agent    => m[6]
					})
					
					$log.push entry
				else
					entry = $log.find {|i| i.ip == m[1] and \
										   i.name == name and \
										   not i.finished}
					entry.finished = true if entry
				end
			rescue StandardError
				$stderr.puts "Ignoring invalid entry (line #{cur})"
			end
			cur += 1
		end
	end
end

"""
	Create a simple summary that includes the name of each video played, the
	number of time the video was started, the number of times it was finished,
	and a list of all unique IPs that played that video.
	
	A list of SummaryItems sorted by hit count is returned.
"""
def simple_summary()
	summary = []
	names = []
	$log.each do |item|
		if not names.member? item.name
			names.push item.name
			summary.push SummaryItem.new({ :name => item.name })
		end
		sumitem = summary.find {|i| i.name == item.name}
		sumitem.hits += 1
		sumitem.full_views += 1 if item.finished
		sumitem.unique_ips.push(item.ip) if not \
									sumitem.unique_ips.member? item.ip
	end
	summary.sort_by { |item| item.hits }.reverse
end

"""
	Print out information about the simple summary returned from the function
	above in an easy to read table.
"""
def print_simple_summary(summary = simple_summary)
	title = "Video Hit Summary"
	labels = ["Name", "Starts", "Stops", "Unique IPs"]
	maxlen = [labels[0].length, labels[1].length,
			  labels[2].length, labels[3].length]
	
	summary.each do |item|
		if item.name.length > maxlen[0]
			maxlen[0] = item.name.length
		end
		if item.hits.to_s.length > maxlen[1]
			maxlen[1] = item.hits.to_s.length
		end
		if item.full_views.to_s.length > maxlen[2]
			maxlen[2] = item.full_views.to_s.length
		end
		if item.unique_ips.length.to_s.length > maxlen[3]
			maxlen[3] = item.unique_ips.length.to_s.length
		end
	end
	width = maxlen.inject {|sum, n| sum + n} + 9
	puts "/" + "=" * (width - 2) + "\\"
	titlespace = (width / 2 - title.length / 2)
	puts "|" + " " * (titlespace - 1) + title + " " * (titlespace - 1) + "|"
	puts "-" * width
	(0..3).each do |num|
		diff = maxlen[num] - labels[num].length
		print labels[num]
		print " " * diff if diff > 0
		print " | " if num != 3
	end
	print "\n"
	puts "-" * width
	summary.each do |item|
		print item.name
		print " " * (maxlen[0] - item.name.length) if \
					(maxlen[0] - item.name.length) > 0
		print " | "
		[item.hits.to_s, item.full_views.to_s, \
			item.unique_ips.length.to_s].each_with_index do |item, i|
			print " " * (maxlen[i + 1] - item.length) if \
											(maxlen[i + 1] - item.length) > 0
			print item
			print " | " if i + 1 != 3
		end
		print "\n"
	end
	puts "\\" + "=" * (width - 2) + "/"
end

"""
	Create a bar chart of the simple summary returned above and save it as
	an SVG file.
"""
def simple_summary_bar_chart(summary = simple_summary, filename = \
							 "summary.svg", count = 10)
	fields = []
	hits = []
	stops = []
	ips = []
	
	# Fill our data values for [count] most hit videos
	current = 0
	summary.each do |item|
		fields.push item.name
		hits.push item.hits
		stops.push item.full_views
		ips.push item.unique_ips.length
		current += 1
		break if current == count
	end
	
	# Create the graph
	graph = SVG::Graph::Bar.new(
		:graph_title => "Top " + current.to_s + " Most Visited Videos",
		:show_graph_title => true,
		:title_font_size => 20,
		:graph_subtitle => "Longer bars mean more hits",
		:show_graph_subtitle => true,
		:show_data_values => false,
		:height => 650,
		:width => 800,
		:fields => fields,
		:scale_integers => true,
		:stack => :side,
		:y_title => "Hits",
		:show_y_title => true,
		:rotate_x_labels => true
	)
	
	graph.add_data(
		:data => hits,
		:title => "Video Hits"
	)
	
	graph.add_data(
		:data => stops,
		:title => "Full Views"
	)
	
	graph.add_data(
		:data => ips,
		:title => "Unique Watchers"
	)
	
	desc = \
	"A summary of which videos are the most popular based on how many times
	they are started."
	
	graph.add_text(685, 130, 110, 300, desc)
	
	# Save the SVG file
	File.open(filename, "w") do |f|
		f.puts graph.burn
	end
end

"""
	Create a nice XHTML/CSS page with the results of the log summary using
	a template file.
"""
def stats_xhtml(summary = simple_summary)
	title = "Video Stats"
	graph = "summary.svg"
	
	def print_summary_rows(summary)
		rows = ""
		alt = false
		summary.each do |item|
			label = "No label" if label == ""
			rows += alt ? '<tr class="alt">' : "<tr>"
			rows += %Q~<td>#{item.name}</td>~
			rows += %Q~<td class="num">#{item.hits}</td>~
			rows += %Q~<td class="num">#{item.full_views}</td>~
			rows += %Q~<td class="num">#{item.unique_ips.length}</td>~
			rows += "</tr>\n"
			alt = !alt
		end
		rows
	end
	
	begin
		xhtml = File.open("xhtml.template") { |f| f.read }
		
		File.open("stats.html", "w") do |f|
			f.puts xhtml.interpolate(binding)
		end
	rescue StandardError
		$stderr.puts "Error opening XHTML template."
	end
end

def usage
	puts $NAME + " by " + $AUTHOR
	puts "Version " + $VERSION + " Copyright " + $COPYRIGHT
	puts "Usage:"
	puts "\t" + $0 + " logfile"
end

# Open the log, load the data, and print a simple summary
if ARGV.length == 0
	usage
	exit
end

if ARGV[0] == "-h" or ARGV[0] == "--help"
	usage
	exit
end

if not File.exist? ARGV[0]
	$stderr.puts "File (#{ARGV[0]}) doesn't exist. Aborting."
	exit
end

read_log ARGV[0]
summary = simple_summary
print_simple_summary summary
simple_summary_bar_chart summary
stats_xhtml summary
